#ifndef MATH_H_
#define MATH_H_

#include <cmath>
#include <string>

namespace arda 
   {

   ////////////////////////////////////////////////////////////////////////////////
   // General notes:
   //
   // As much as possible is defined inline for performance.
   // Using typedef'ed arrays for vectors and matrices necessitates a lot of 
   // redundant function names: scale2, scale3, scale4, for example.
   // This is, however, perfectly in line with the implementation of OpenGL and
   // allows the library to maintain optimal memory usage and performance. (I hope!)
   // In any event, I think I could do far worse than to emulate the OpenGL
   // working group - a group of people who have studied this for years and are
   // experts in the field.

   ////////////////////////////////////////////////////////////////////////////////
   // Vectors

   namespace Vector 
      {

      typedef int Vector2i[2];
      typedef float Vector2f[2];
      typedef double Vector2d[2];

      typedef int Vector3i[3];
      typedef float Vector3f[3];
      typedef double Vector3d[3];

      typedef int Vector4i[4];
      typedef float Vector4f[4];
      typedef double Vector4d[4];

      //////////////////////////////////////////////////////////////////////////
      // vectostr*(Vector v)
      //	Returns a C string representation of the vector.
      //	This is one of the few non-inline Vector functions.

      std::string & vectostr2(Vector2i v, std::string & s);
      std::string & vectostr2(Vector2f v, std::string & s);
      std::string & vectostr2(Vector2d v, std::string & s);
      std::string & vectostr3(Vector3i v, std::string & s);
      std::string & vectostr3(Vector3f v, std::string & s);
      std::string & vectostr3(Vector3d v, std::string & s);
      std::string & vectostr4(Vector4i v, std::string & s);
      std::string & vectostr4(Vector4f v, std::string & s);
      std::string & vectostr4(Vector4d v, std::string & s);
      
      //////////////////////////////////////////////////////////////////////////
      // scale*(Vector v, scalar s)
      //	Multiply a vector by a scalar.
      //	Returns v so that the result may be passed to other functions.

      inline int * scale2(Vector2i v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline int * scale2(Vector2i v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline int * scale2(Vector2i v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }

      inline float * scale2(Vector2f v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline float * scale2(Vector2f v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline float * scale2(Vector2f v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }

      inline double * scale2(Vector2d v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline double * scale2(Vector2d v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }
      inline double * scale2(Vector2d v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; return v; }


      inline int * scale3(Vector3i v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline int * scale3(Vector3i v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline int * scale3(Vector3i v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
 
      inline float * scale3(Vector3f v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline float * scale3(Vector3f v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline float * scale3(Vector3f v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }

      inline double * scale3(Vector3d v, const int s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline double *  scale3(Vector3d v, const float s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }
      inline double * scale3(Vector3d v, const double s)
	 { v[0] = v[0] * s ; v[1] = v[1] * s; v[2] = v[2] * s; return v; }


      inline int * scale4(Vector4i v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline int * scale4(Vector4i v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline int * scale4(Vector4i v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }

      inline float * scale4(Vector4f v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline float * scale4(Vector4f v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline float * scale4(Vector4f v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }

      inline double * scale4(Vector4d v, const int s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline double * scale4(Vector4d v, const float s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }
      inline double * scale4(Vector4d v, const double s) 
	 { v[0] = v[0] * s ; v[1] = v[1] * s;  v[2] = v[2] * s;  v[3] = v[3] * s; return v; }


      //////////////////////////////////////////////////////////////////////////
      // dot*(Vector v1, Vector v2)
      //	The dot product of two vectors, which must be the same type
      //	Returns a long for int vectors, a double for floating point vectors.

      inline long dot2(Vector2i v1, Vector2i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1]; }
      inline double dot2(Vector2f v1, Vector2f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1]; }
      inline double dot2(Vector2d v1, Vector2d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1]; }

      inline long dot3(Vector3i v1, Vector3i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2]; }
      inline double dot3(Vector3f v1, Vector3f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2]; }
      inline double dot3(Vector3d v1, Vector3d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]; }

      inline long dot4(Vector4i v1, Vector4i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2] + (long) v1[3]*v2[3]; }
      inline double dot4(Vector4f v1, Vector4f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2] + (double) v1[3]*v2[3]; }
      inline double dot4(Vector4d v1, Vector4d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2] + v1[3]*v2[3]; }


      //////////////////////////////////////////////////////////////////////////
      // length*(Vector v)
      //	Get the length of a vector.
      //	Returns a float for float for float vectors, and a double for 
      //	int and double vectors.

      inline double length2(Vector2i v)
	 { return sqrt((dot2(v, v))); }
      inline float length2(Vector2f v)
	 { return sqrtf((dot2(v, v))); }
      inline double length2(Vector2d v)
	 { return sqrt((dot2(v, v))); }

      inline double length3(Vector3i v)
	 { return sqrt((dot3(v, v))); }
      inline float length3(Vector3f v)
	 { return sqrtf((dot3(v, v))); }
      inline double length3(Vector3d v)
	 { return sqrt((dot3(v, v))); }

      inline double length4(Vector4i v)
	 { return sqrt((dot4(v, v))); }
      inline float length4(Vector4f v)
	 { return sqrtf((dot4(v, v))); }
      inline double length4(Vector4d v)
	 { return sqrt((dot4(v, v))); }


      //////////////////////////////////////////////////////////////////////////
      // add*(Vector v1, Vector v2, Vector vres)
      //	Add two vectors together.
      //        Returns vres so that the result may be used in further 
      //	calculations.

      inline int * add2(Vector2i v1, Vector2i v2, Vector2i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }
      inline float * add2(Vector2f v1, Vector2f v2, Vector2f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }
      inline double * add2(Vector2d v1, Vector2d v2, Vector2d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; return vres; }

      inline int * add3(Vector3i v1, Vector3i v2, Vector3i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; return vres; }
      inline float * add3(Vector3f v1, Vector3f v2, Vector3f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2];  return vres; }
      inline double * add3(Vector3d v1, Vector3d v2, Vector3d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2];  return vres; }

      inline int * add4(Vector4i v1, Vector4i v2, Vector4i vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres;}
      inline float * add4(Vector4f v1, Vector4f v2, Vector4f vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres; }
      inline double * add4(Vector4d v1, Vector4d v2, Vector4d vres)
	 { vres[0] = v1[0] + v2[0]; vres[1] = v1[1] + v2[1]; vres[2] = v1[2] + v2[2]; vres[3] = v1[3] + v2[3]; return vres; }

      //////////////////////////////////////////////////////////////////////////
      // normalize*(Vector v)
      //	Normalizes the vecor.  Only defined for floating point vectors.
      //	Returns the vector.

      inline float * normalize2(Vector2f v)
	 { float l = length2(v); v[0] /= l; v[1] /= l; return v; }
      inline double * normalize2(Vector2d v)
	 { double l = length2(v); v[0] /= l; v[1] /= l; return v; }

      inline float * normalize3(Vector3f v)
	 { float l = length3(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }
      inline double * normalize3(Vector3d v)
	 { double l = length3(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }

      inline float * normalize4(Vector4f v)
	 { float l = length4(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }
      inline double * normalize4(Vector4d v)
	 { double l = length4(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }
   }

} // namespace arda

#endif
